﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Parkhaus;
using System.Threading;

namespace ParkhausGDI
{
    public partial class FormParkhaus : Form, ICallbackHandler
    {
        #region Variables
        private delegate void delegateParking(int countPlaces, List<Auto> _parkingList);
        private delegateParking updateParkingDelegate;

        Parkhaus.Parkhaus _parkhaus;
        Random _rnd = new Random();
        protected object _locker = new object();
        List<Auto> _parkingList;
        List<Thread> _RunningCars = new List<Thread>();
        int _freePlaces;
        int _maxPlaces;
        Queue<Auto> _changes = new Queue<Auto>(10);
        #endregion

        #region Constructors
        public FormParkhaus()
        {
            InitializeComponent();
            // optimiertes Zeichnen
            this.SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.OptimizedDoubleBuffer | ControlStyles.ResizeRedraw, true);           
            this.UpdateStyles();
            updateParkingDelegate = new delegateParking(UpdateParking);
            _parkingList = new List<Auto>();

            _maxPlaces = 10;
            _parkhaus = new Parkhaus.Parkhaus(_maxPlaces, this);
            _freePlaces = _maxPlaces;

            for (int i = 1; i <= 40; i++)
            {
                Color c = System.Drawing.Color.FromArgb(_rnd.Next(256), _rnd.Next(256), _rnd.Next(256));

                Auto auto = new Auto("Auto " + i.ToString("00"), c, _parkhaus, ref _changes);
                Thread t = new Thread(auto.Drive);
                t.Start();
                _RunningCars.Add(t); // In Threadliste füllen, damit alle Threads wieder beendet werden können
            }
        }
        #endregion

        #region Painting
        private void UpdateParking(int countPlaces, List<Auto> parkingList)
        {
            // Da Threadübergreifender Aufruf abfrage ob Invoke nötig ist
            if (InvokeRequired)
            {
                Invoke(updateParkingDelegate, new object[] { countPlaces, parkingList });
            }
            else
            {
                lock (_locker)
                {
                    this.Invalidate(); // damit neu gezeichnet wird
                    _freePlaces = countPlaces; // Anzahl freie Plätze
                    _parkingList = parkingList;
                }
            }
        }

        private void DrawCar(Graphics g, int pos, Auto car)
        {
            int carWidth = 18;
            int halfHeight = (int)(this.Height / 2);
            int halfWidth = (int)(this.Width / 2);
            int SevenPointThreeWidth = (int)(this.Width / 7.3);
            int OnePointSixHeight = (int)(this.Height / 1.63);
            int x1 = (halfWidth - SevenPointThreeWidth / 2);
            int y1 = this.Height - OnePointSixHeight;
            int x2 = (halfWidth - SevenPointThreeWidth / 6);
            int y2 = this.Height - (int)(this.Height / 1.83);
            int x3 = (halfWidth + SevenPointThreeWidth / 2);

            int x = 0;
            int y = 0;
            // Fahrzeug position
            switch (pos)
            {
                case 0:
                    x = x1 + (x2 - x1) + (x2 - x1) / 2 - carWidth / 2;
                    y = y1 + 3 * (y2 - y1) + (y2 - y1) / 2;
                    break;
                case 1:
                case 2:
                case 3:
                    x = x1 + (x2 - x1) / 2 - carWidth / 2;
                    y = y1 + (pos - 1) * (y2 - y1) + (y2 - y1) / 2;
                    break;
                case 4:
                case 5:
                case 6:
                    x = x1 + (x2 - x1) + (x2 - x1) / 2 - carWidth / 2;
                    y = y1 + (pos - 4) * (y2 - y1) + (y2 - y1) / 2;
                    break;
                case 7:
                case 8:
                case 9:
                    x = x1 + 2 * (x2 - x1) + (x2 - x1) / 2 - carWidth / 2;
                    y = y1 + (pos - 7) * (y2 - y1) + (y2 - y1) / 2;
                    break;
                default:
                    break;
            }

            // Räder
            g.FillEllipse(Brushes.Black, new Rectangle(x + 2, y + 10, 4, 4));
            g.FillEllipse(Brushes.Black, new Rectangle(x + carWidth - 4, y + 10, 4, 4));
            // Karosserie
            SolidBrush sb = new SolidBrush(car.color);
            g.FillRectangle(sb, new Rectangle(x + 1, y + 3, carWidth, 6)); // Rumpf
            g.FillRectangle(sb, new Rectangle(x + carWidth / 4 + carWidth / 4, y + 7, carWidth / 3, 5)); // Unterboden
            g.FillEllipse(sb, new Rectangle(x + carWidth / 3, y, carWidth / 2, carWidth / 2)); // Dach
            // Auto-Bezeichnung
            float fontSize = (float)(this.Height / 54.87);
            Font f = new Font("Arial", fontSize);
            String carNumber = car.name.Substring(car.name.IndexOf(' ', 0) + 1);
            SizeF fontMesurementCarNumber = g.MeasureString(carNumber, f);
            g.DrawString(carNumber, f, Brushes.Black, x + (carWidth / 2) - fontMesurementCarNumber.Width, y + 5);
        }

        protected override void OnPaintBackground(PaintEventArgs e)
        {
            base.OnPaintBackground(e);
            e.Graphics.Clear(this.BackColor);
            e.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
            int halfHeight = (int)(this.Height / 2);
            int halfWidth = (int)(this.Width / 2);
            int SevenPointThreeWidth = (int)(this.Width / 7.3);
            int OnePointSixHeight = (int)(this.Height / 1.63);
            // Hügel
            e.Graphics.FillEllipse(Brushes.Green, new Rectangle(0, halfHeight, this.Width, halfHeight));
            e.Graphics.FillEllipse(Brushes.Green, new Rectangle(-halfWidth, halfHeight, this.Width, halfHeight));
            e.Graphics.FillEllipse(Brushes.Green, new Rectangle(halfWidth, halfHeight, this.Width, halfHeight));
            // Parkhaus
            // Parkhaus Ebenen
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth - SevenPointThreeWidth / 2, this.Height - (int)(this.Height / 2.93), SevenPointThreeWidth, 5));
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth - SevenPointThreeWidth / 2, this.Height - (int)(this.Height / 2.44), SevenPointThreeWidth, 5));
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth - SevenPointThreeWidth / 2, this.Height - (int)(this.Height / 2.09), SevenPointThreeWidth, 5));
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth - SevenPointThreeWidth / 2, this.Height - (int)(this.Height / 1.83), SevenPointThreeWidth, 5));
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth - SevenPointThreeWidth / 2, this.Height - OnePointSixHeight, SevenPointThreeWidth, 5));
            // Parkhaus Wände
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth - SevenPointThreeWidth / 2, this.Height - OnePointSixHeight, 5, (this.Height - (int)(this.Height / 2.93)) - (this.Height - OnePointSixHeight)));
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth - SevenPointThreeWidth / 6, this.Height - OnePointSixHeight, 5, (this.Height - (int)(this.Height / 2.93)) - (this.Height - OnePointSixHeight)));
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth + SevenPointThreeWidth / 6, this.Height - OnePointSixHeight, 5, (this.Height - (int)(this.Height / 2.93)) - (this.Height - OnePointSixHeight)));
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth + SevenPointThreeWidth / 2, this.Height - OnePointSixHeight, 5, (this.Height - (int)(this.Height / 2.93)) - (this.Height - OnePointSixHeight) + 4));

            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth - SevenPointThreeWidth / 2, this.Height - (int)(this.Height / 2.44), (halfWidth - SevenPointThreeWidth / 6) - (halfWidth - SevenPointThreeWidth / 2), (this.Height - (int)(this.Height / 2.93)) - (this.Height - (int)(this.Height / 2.44))));
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(halfWidth + SevenPointThreeWidth / 6, this.Height - (int)(this.Height / 2.44), (halfWidth - SevenPointThreeWidth / 6) - (halfWidth - SevenPointThreeWidth / 2), (this.Height - (int)(this.Height / 2.93)) - (this.Height - (int)(this.Height / 2.44))));

            // Schriften
            float fontSize = (float)(this.Height / 54.87);
            Font f = new Font("Arial", fontSize);
            SizeF fontMesurementParkhaus = e.Graphics.MeasureString("Parkhaus", f);
            SizeF fontMesurementEntry = e.Graphics.MeasureString("Eingang freie Plätze:" + _freePlaces.ToString(), f);
            SizeF fontMesurementExit = e.Graphics.MeasureString("Ausgang", f);

            // Schrift -> Parkhaus
            Rectangle rParkhaus = new Rectangle((int)(halfWidth - (fontMesurementParkhaus.Width / 2)), this.Height - (int)(this.Height / 1.53), (int)(fontMesurementParkhaus.Width), (int)(fontMesurementParkhaus.Height));
            e.Graphics.FillRectangle(Brushes.Brown, rParkhaus);
            e.Graphics.FillRectangle(Brushes.Brown, rParkhaus.X, rParkhaus.Y + rParkhaus.Height, 5, 10);
            e.Graphics.FillRectangle(Brushes.Brown, rParkhaus.X + rParkhaus.Width - 5, rParkhaus.Y + rParkhaus.Height, 5, 10);
            e.Graphics.DrawString("Parkhaus", f, Brushes.White, new PointF(rParkhaus.X, rParkhaus.Y));

            // Schrift -> Eingang
            Rectangle rEntry = new Rectangle((int)((halfWidth - SevenPointThreeWidth / 2) - fontMesurementEntry.Width), (int)(this.Height - (int)(this.Height / 2.93)), (int)(fontMesurementEntry.Width), (int)(fontMesurementEntry.Height));
            e.Graphics.FillRectangle(Brushes.Brown, rEntry);
            e.Graphics.FillRectangle(Brushes.Brown, rEntry.X, rEntry.Y + rEntry.Height, 5, 10);
            e.Graphics.FillRectangle(Brushes.Brown, rEntry.X + rEntry.Width - 5, rEntry.Y + rEntry.Height, 5, 10);
            switch (_freePlaces)
            {
                case 0:
                    e.Graphics.DrawString("Eingang freie Plätze:" + _freePlaces.ToString(), f, Brushes.RosyBrown, new PointF(rEntry.X, rEntry.Y));
                    break;
                default:
                    e.Graphics.DrawString("Eingang freie Plätze:" + _freePlaces.ToString(), f, Brushes.YellowGreen, new PointF(rEntry.X, rEntry.Y));
                    break;
            }

            // Schrift -> Ausgang
            Rectangle rExit = new Rectangle((int)(halfWidth + SevenPointThreeWidth / 2) + 5, (int)(this.Height - (int)(this.Height / 2.93)), (int)(fontMesurementExit.Width), (int)(fontMesurementExit.Height));
            e.Graphics.FillRectangle(Brushes.Brown, rExit);
            e.Graphics.FillRectangle(Brushes.Brown, rExit.X, rExit.Y + rExit.Height, 5, 10);
            e.Graphics.FillRectangle(Brushes.Brown, rExit.X + rExit.Width - 5, rExit.Y + rExit.Height, 5, 10);
            e.Graphics.DrawString("Ausgang", f, Brushes.White, new PointF(rExit.X, rExit.Y));

            // Copyright
            string copyright = "Copyright Bruno Bieri TA.I.0801";
            SizeF fontMesurementCopyright = e.Graphics.MeasureString(copyright, f);
            e.Graphics.DrawString(copyright, f, Brushes.White, new PointF(this.ClientRectangle.Width - fontMesurementCopyright.Width, this.ClientRectangle.Height - (1 * fontMesurementCopyright.Height)));
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            // Autos zeichnen
            if (_parkingList.Count > 0)
            {
                for (int i = 0; i < _parkingList.Count; i++)
                {
                    if (!(_parkingList[i] == null))
                    {
                        DrawCar(e.Graphics, i, _parkingList[i]);
                    }
                }
            }

            // Status
            int halfWidth = (int)(this.Width / 2);
            int SixWidth = (int)(this.Width / 4);
            int OnePointSixHeight = (int)(this.Height / 1.63);
            int x = halfWidth + SixWidth;
            int y = this.Height - OnePointSixHeight;

            float fontSize = (float)(this.Height / 54.87);
            Font f = new Font("Arial", fontSize);
            SolidBrush sb = new SolidBrush(Color.White);
            e.Graphics.FillRectangle(Brushes.Black, new Rectangle(x - 4, y - 4, (int)(e.Graphics.MeasureString("Auto 00: ausgefahren", f).Width + 8), (int)(10 * (fontSize + 4)) + 12));
            if (_changes.Count > 0)
            {
                for (int i = _changes.Count - 1; i > -1; i--)
                {
                    sb.Color = _changes.ElementAt(i).color;
                    e.Graphics.DrawString(_changes.ElementAt(i)._inParking ? _changes.ElementAt(i).name + ": eingefahren" : _changes.ElementAt(i).name + ": ausgefahren", f, sb, new PointF(x, y + i * (fontSize + 4)));
                }
            }
        }
        #endregion

        #region ICallbackHandler Members

        public new void Handle(object res)
        {
            // Annahme: Uebergebenes Objekt ist ein Array!
            object[] paras = (object[])res;
            UpdateParking((int)(paras[0]), (List<Auto>)(paras[1]));
        }

        #endregion

        private void FormParkhaus_FormClosing(object sender, FormClosingEventArgs e)
        {
            for (int i = 0; i < _RunningCars.Count; i++)
            {
                _RunningCars[i].Abort();
            }
        }
    }
}
